import asyncio
from typing import List

import httpx

from argos.checks import get_registered_check
from argos.logging import logger
from argos.schemas import AgentResult, SerializableException, Task


async def complete_task(http_client: httpx.AsyncClient, task: dict) -> dict:
    try:
        task = Task(**task)
        check_class = get_registered_check(task.check)
        check = check_class(http_client, task)
        result = await check.run()
        status = result.status
        context = result.context

    except Exception as e:
        status = "error"
        context = SerializableException.from_exception(e)
        msg = f"An exception occured when running {task}. {e.__class__.__name__} : {e}"
        logger.error(msg)
    return AgentResult(task_id=task.id, status=status, context=context)


async def post_results(
    http_client: httpx.AsyncClient, server: str, results: List[AgentResult]
):
    data = [r.model_dump() for r in results]
    response = await http_client.post(f"{server}/api/results", json=data)

    if response.status_code == httpx.codes.CREATED:
        logger.error(f"Successfully posted results {response.json()}")
    else:
        logger.error(f"Failed to post results: {response.read()}")
    return response


async def get_and_complete_tasks(http_client, server, max_tasks):
    # Fetch the list of tasks
    response = await http_client.get(f"{server}/api/tasks")

    if response.status_code == httpx.codes.OK:
        # XXX Maybe we want to group the tests by URL ? (to issue one request per URL)
        data = response.json()
        logger.info(f"Received {len(data)} tasks from the server")

        tasks = []
        for task in data:
            tasks.append(complete_task(http_client, task))

        if tasks:
            results = await asyncio.gather(*tasks)
            await post_results(http_client, server, results)
            return True
        else:
            logger.error("Got no tasks from the server.")
            return False
    else:
        logger.error(f"Failed to fetch tasks: {response.read()}")
        return False


async def run_agent(server: str, auth: str, max_tasks: int, wait_time: int):
    headers = {"Authorization": f"Bearer {auth}"}
    async with httpx.AsyncClient(headers=headers) as http_client:
        while True:
            retry_now = await get_and_complete_tasks(http_client, server, max_tasks)
            if not retry_now:
                logger.error(f"Waiting {wait_time} seconds before next retry")
                await asyncio.sleep(wait_time)
