from collections import defaultdict
from datetime import datetime
from urllib.parse import urljoin

from sqlalchemy import desc, func
from sqlalchemy.orm import Session, aliased

from argos import schemas
from argos.logging import logger
from argos.server.models import Result, Task


async def list_tasks(db: Session, agent_id: str, limit: int = 100):
    """List tasks and mark them as selected"""
    tasks = (
        db.query(Task)
        .filter(
            Task.selected_by == None,  # noqa: E711
            ((Task.next_run <= datetime.now()) | (Task.next_run == None)),  # noqa: E711
        )
        .limit(limit)
        .all()
    )

    now = datetime.now()
    for task in tasks:
        task.selected_at = now
        task.selected_by = agent_id
    db.commit()
    return tasks


async def get_task(db: Session, id: int) -> Task:
    return db.get(Task, id)


async def create_result(db: Session, agent_result: schemas.AgentResult):
    result = Result(
        submitted_at=datetime.now(),
        status=agent_result.status,
        context=agent_result.context,
        task_id=agent_result.task_id,
    )
    db.add(result)
    return result


async def count_tasks(db: Session, selected=False):
    query = db.query(Task)
    if selected:
        query = query.filter(Task.selected_by is not None)
    else:
        query = query.filter(Task.selected_by is None)

    return query.count()


async def count_results(db: Session):
    return db.query(Result).count()


async def update_from_config(db: Session, config: schemas.Config):
    for website in config.websites:
        domain = str(website.domain)
        frequency = website.frequency or config.general.frequency

        for p in website.paths:
            url = urljoin(domain, str(p.path))
            for check_key, expected in p.checks:
                # Check the db for already existing tasks.
                existing_task = (
                    db.query(Task)
                    .filter(
                        Task.url == url,
                        Task.check == check_key,
                        Task.expected == expected,
                    )
                    .scalar()
                )
                if existing_task and frequency != existing_task.frequency:
                    existing_task.frequency = frequency

                if not existing_task:
                    task = Task(
                        domain=domain,
                        url=url,
                        check=check_key,
                        expected=expected,
                        frequency=frequency,
                    )
                    logger.debug(f"Adding a new task in the db: {task}")
                    db.add(task)
                else:
                    msg = f"Skipping db task creation for {url=}, {check_key=}, {expected=}, {frequency=}."
                    logger.debug(msg)
    db.commit()


async def get_severity_counts(db):
    # Get the last result of each task
    subquery = (
        db.query(Result.task_id, func.max(Result.id).label("max_result_id"))
        .group_by(Result.task_id)
        .subquery()
    )

    # Join this back to get full result rows, and group by status
    query = (
        db.query(Result.severity, func.count(Result.id).label("count"))
        .join(subquery, Result.id == subquery.columns.max_result_id)
        .group_by(Result.severity)
    )

    # Execute the query and fetch the results
    task_counts_by_severity = query.all()
    return task_counts_by_severity


async def get_domains_with_severities(db):
    tasks = db.query(Task).group_by(Task.domain).all()

    domains = defaultdict(list)
    for task in tasks:
        domains[task.domain].append(task.severity)

    def _max_severity(severities):
        severity_level = {"ok": 1, "warning": 2, "critical": 3}
        return max(severities, key=severity_level.get)

    return {key: _max_severity(value) for key, value in domains.items()}
